/******************************************************************************
 * FILE:
 *    shared_main.c
 *
 * DESCRIPTION:
 *    Multi-core LED Blinking skeleton, where each core drives its own LED
 *
 * FILE HISTORY:
 *    V1.0    17.10.2017     RO
 *       > Initial version
 *    V2.0    24.01.2018     RO
 *       > Header file update
 *       > BSP folder restructure needs include update
 *       > common_main renamed to shared_name to better reflect entry point
 *         multi-core characteristic
 *
 ******************************************************************************
 * Copyright (C) 2015-2018 HighTec EDV-Systeme GmbH.  All rights reserved.
 * This is proprietary software. Any use of the software requires a written
 * license agreement with HighTec EDV-Systeme GmbH. Please contact for
 * detailed license information: info@hightec-rt.com.
 *****************************************************************************/


/* ============================================================================
 *                           INCLUDE SECTION
 * ==========================================================================*/

/* BSP support package */
#include "bsp.h"
/* SAR ADC EXAMPLE*/
#include "saradc.h"

/* ============================================================================
 *                           EXTERNAL SYMBOL DECLARATION
 * ==========================================================================*/


/* ============================================================================
 *                           LOCAL DEFINES & MACROS
 * ==========================================================================*/

/* BSP Timer Period Base */
#define TIMER_BASE_PERIOD         250 * BSP_TICKS_1MS


/* ============================================================================
 *                           LOCAL SYMBOL DECLARATION
 * ==========================================================================*/

/* Application Entry Point declaration */
void shared_main(int coreId);

/* Interrupt handler for BSP Timer channels */
void TimerMulticoreIsrHandler(void);


/* ============================================================================
 *                           LOCAL DATA
 * ==========================================================================*/


/* ============================================================================
 *                           ISR FUNCTIONS
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: TimerMulticoreIsrHandler()
 *    BSP Timer ISR Handler
 * ----------------------------------------------------------------------------
 */
void TimerMulticoreIsrHandler(void)
{
	uint32_t coreId = bsp_uc_GetCurrentCoreId();

	/* SARADC  EXAMPLE */
	/* start ADC channel scan */
	saradc_StartScan();

	bsp_pin_Toggle(coreId);
	bsp_timer_ClearChannelIsrFlag(coreId);
}


/* ============================================================================
 *                           API FUNCTIONS
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: shared_main()
 *    Shared entry point if there are more core present in uC derivative.
 *    Current core entering the function is passed as parameter.
 *    shared_main() entry point is called by crt0 startup code when it finishes
 *    'C' runtime environment initialization.
 * --------------------------------------------------------------------------*/
void shared_main(int coreId)
{
	/* Each core initialization - different timer channel and LED IO */
	bsp_timer_StopChannel(coreId);
	bsp_timer_ReloadChannel(coreId, (coreId + 1) * TIMER_BASE_PERIOD);
	bsp_timer_EnableChannelIsr(coreId, 5, TimerMulticoreIsrHandler);
	bsp_pin_EnableBidir(coreId, 0);

	/* SARADC  EXAMPLE */
    /* Initialize SARADC B */
	saradc_Init();

	bsp_timer_StartChannel(coreId);

	/* Dummy Infinitive loop
	 * LED blinking done by interrupt (shared handler) on each core */
	while(1);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: Crt0PreInit()
 *    Crt0 Hook before memory clear & copy operation take place
 * INPUT:
 *    coreId : current execution core identifier
 * --------------------------------------------------------------------------*/
void Crt0PreInit(uint32_t coreId)
{
	/* Default initialization provided by BSP */
	bsp_crt0_PreInit(coreId);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: Crt0PostInit()
 *    Crt0 Hook after complete crt0 initialization, just before
 *    shared_main() entry point is called
 * INPUT:
 *    coreId : current execution core identifier
 * --------------------------------------------------------------------------*/
void Crt0PostInit(uint32_t coreId)
{
	/* Default initialization provided by BSP */
	bsp_crt0_PostInit(coreId);
}
