/******************************************************************************
 * FILE
 *    uc_lld.c
 *
 * UC ARCHITECTURE:
 *    PowerPC
 *
 * UC FAMILY:
 *    SPC57
 *
 * DESCRIPTION
 *    Low Level Register access implementation through Modules' base addresses
 *    and register access macro functions in uc_regdefs.h.
 *
 * FILE HISTORY:
 *    V1.0    23.09.2017     RO
 *       > Initial version
 *    V2.0    25.01.2018     RO
 *       > File name rename and file header update
 *
 ******************************************************************************
 * Copyright (C) 2015-2018 HighTec EDV-Systeme GmbH.  All rights reserved.
 * This is proprietary software. Any use of the software requires a written
 * license agreement with HighTec EDV-Systeme GmbH. Please contact for
 * detailed license information: info@hightec-rt.com.
 *****************************************************************************/

/* Include its own API interface and types */
#include "uc_lld.h"


/* ============================================================================
 * LLD DEFINES
 * ==========================================================================*/

/* The only ME mode supported by BSP LLD driver */
#define UC_MODE_DRUN         3


/* ============================================================================
 * LLD API
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_me_ChangeMode2DRUN()
 *    It executes Mode Entry transition to DRUN. Everything that is linked
 *    to DRUN mode gets activated.
 * NOTE:
 *    Function waits in the loop for transition and change of mode to finish
 *    without timeout.
 * --------------------------------------------------------------------------*/
void uc_me_ChangeMode2DRUN(void)
{
	UC_ME_MCTL_REG = ((UC_MODE_DRUN & (uint32_t)0x0F) << 28) | (uint32_t) 0x5AF0;
	UC_ME_MCTL_REG = ((UC_MODE_DRUN & (uint32_t)0x0F) << 28) | (uint32_t) 0xA50F;
	while (UC_ME_GS_REG & UC_ME_GS_MTRANS_BIT);
    while (MC_ME_GS_GET_MODE(UC_ME_GS_REG) != UC_MODE_DRUN);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_cgm_EnableXOSC()
 *    It enables an internal oscillator block in DRUN mode. After mode
 *    transition it waits for external clock stabilization in infinitive loop.
 * --------------------------------------------------------------------------*/
void uc_cgm_EnableXOSC(void)
{
	UC_ME_DRUN_MC_REG |= UC_ME_MODE_XOSC_BIT;
	uc_me_ChangeMode2DRUN();
	while (!(UC_ME_GS_REG & UC_ME_MODE_XOSC_BIT));
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_cgm_EnablePLL()
 *    It enables PLL0 in DRUN mode with required parameters. After transition
 *    it waits for PLL0 stabilization in infinitive loop.
 * INPUTS:
 *    instance : selects which PLL(x) to initialize if asked for
 *    prediv   : Input divider (PLL0 yes, PLL1 no)
 *    mfd      : VCO feedback divider (both PLLs)
 *    rfdphi   : main output (both PLLs)
 *    rfdphi1  : secondary output PLL0->PLL1 (PLL1 no)
 * --------------------------------------------------------------------------*/
void uc_cgm_EnablePLL(uint32_t instance, uint32_t prediv, uint32_t mfd, uint32_t rfdphi, uint32_t rfdphi1)
{
	uint32_t pll_bit = (instance) ? UC_ME_MODE_PLL1_BIT : UC_ME_MODE_PLL0_BIT;
	vuint32_t* reg_dv_base = (instance) ? UC_PLL1_DV_REG_BASE : UC_PLL0_DV_REG_BASE;
	uint32_t pll_dv = (instance) ? UC_PLL1_DV_VAL(mfd, rfdphi) :
			                       UC_PLL0_DV_VAL(prediv, mfd, rfdphi, rfdphi1);

	UC_ME_DRUN_MC_REG |= pll_bit;
	*reg_dv_base = pll_dv;
	uc_me_ChangeMode2DRUN();
	while (!(UC_ME_GS_REG & pll_bit));
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_cgm_SwitchSystemClock()
 *    Selects requested clock input for SYSTEM clock and makes
 *    a mode transition. This operation shall be executed when the requested
 *    input clock is stable and System clock dividers are set properly.
 * INPUT:
 *    sysclk : clock input from fixed set
 * --------------------------------------------------------------------------*/
void uc_cgm_SwitchSystemClock(uint32_t sysclk)
{
	UC_ME_SET_SYSCLK(UC_ME_DRUN_MC_REG, sysclk);
	uc_me_ChangeMode2DRUN();
	while (UC_ME_GS_GET_SYSCLK(UC_ME_GS_REG) != sysclk);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_cgm_SetScClockDiv()
 *    Setting of System clock dividers. Their values depend on System clock
 *    frequency. Their default (reset) values are set to safe values.
 *    No need to change it when running on max clock.
 * INPUTS:
 *    dc  : DC divider selection
 *    div : divider value (0=disable, !some cannot be disabled)
 * --------------------------------------------------------------------------*/
void uc_cgm_SetScClockDiv(uint32_t dc, uint32_t div)
{
	*(UC_CGM_SC_DC_REG_BASE + dc) = UC_CGM_SET_DC_VAL(div);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_me_EnableAllPeripheralsInDRUN()
 *    Enables DRUN mode for RUN0 configuration which is default configuration
 *    selected in peripheral RUN configuration. The consequence is that all
 *    peripherals will be enabled after the transition to DRUN is executed.
 * --------------------------------------------------------------------------*/
void uc_me_EnableAllPeripheralsInDRUN(void)
{
	UC_ME_RUN_PC0_REG |= UC_ME_RUNPC_DRUN_BIT;
	uc_me_ChangeMode2DRUN();
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_cgm_SelectAuxClock()
 *    Sets requested input clock to selected AUX Bridge.
 * INPUTS:
 *    aux   : AUX Bridge selection
 *    inclk : clock selection from fixed set
 * --------------------------------------------------------------------------*/
void uc_cgm_SelectAuxClock(uint32_t aux, uint32_t inclk)
{
	UC_CGM_AC_REG(aux) = UC_CGM_AC_SET_SELCTL_VAL(inclk);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_cgm_SetAuxClockDiv()
 *    Sets requested divider behind requested AUX Bridge.
 * INPUTS:
 *    aux : AUX Bridge selection (where DC divider belongs)
 *    dc  : Divider selection
 *    div : divider value (0=disable, N=divide by N)
 * NOTE:
 *    It does not support 'fmt' extended divider
 * --------------------------------------------------------------------------*/
void uc_cgm_SetAuxClockDiv(uint32_t aux, uint32_t dc, uint32_t div)
{
	UC_CGM_AC_DC_REG(aux, dc) = UC_CGM_SET_DC_VAL(div);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_core_Start()
 *    It starts requested core in DRUN mode immediately after the transition.
 *    It configures its reset vector from where to start. Enabled core goes
 *    through a reset.
 * INPUTS:
 *    core_nb      : [0..n] number of requested core
 *    reset_vector : pointer to function address where the Core will start
 *                   after the reset
 * --------------------------------------------------------------------------*/
void uc_core_PrepareStart(uint32_t core_nb, uint32_t* reset_vector)
{
	/* configure Core start registers */
	*(UC_ME_CCTL_BASE + core_nb) = UC_ME_CCTL_ALL_RUN_BIT;
	*(UC_ME_CADDR_BASE + core_nb) = (((uint32_t) reset_vector & 0xFFFFFFFC) | (uint32_t) 0x1);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_siul_EnableInput()
 *    Configures requested pin to Input mode with option to enable
 *    pull-up or pull-down
 * INPUTS:
 *    pad_nb          : number of the PAD (number can be found in PAD excel
 *                      sheet attached to RM)
 *    pullup_enable   : enable/disable pull-up
 *    pulldown_enable : enable/disable pull-down
 * --------------------------------------------------------------------------*/
void uc_siul_EnableInput(uint32_t pad_nb, uint32_t pullup_enable, uint32_t pulldown_enable)
{
	UC_SIUL_MSCR_IO_SET_INPUT(*(UC_SIUL_MCSR_IO_BASE + pad_nb), pulldown_enable, pullup_enable);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_siul_EnableOutput()
 *    Configures requested pin to output mode with desired characteristics.
 * INPUTS:
 *    pad_nb : number of the PAD (number can be found in PAD excel sheet
 *             attached to RM)
 *    oert   : edge output rate from fixed set (not all strengths are
 *             supported on every pin)
 *    type   : output PAD type from fixed set (i.e open-drain, Push-pull ...)
 *    source : selection of Alternate function to drive the output
 *             (check excel in RM)
 * NOTE:
 *    Input mode is disabled in this function. User must enable it after
 *    the call, if he needs to read back the value
 * --------------------------------------------------------------------------*/
void uc_siul_EnableOutput(uint32_t pad_nb, uint32_t oerc, uint32_t type, uint32_t source)
{
	UC_SIUL_MSCR_IO_SET_OUTPUT(*(UC_SIUL_MCSR_IO_BASE + pad_nb), oerc, type, source);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_siul_SetGPIO()
 *    Sets value to be driven to output PAD configured in SW driven mode (sss=0)
 * INPUTS:
 *    pad_nb : number of the PAD (number can be found in PAD excel sheet
 *             attached to RM)
 *    value  : one bit value 0/1
 * --------------------------------------------------------------------------*/
void uc_siul_SetGPIO(uint32_t pad_nb, uint8_t value)
{
	*(UC_SIUL_GPDO_BASE + pad_nb) = value & (uint8_t)0x01;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_siul_GetGPIO()
 *    Reads value from the PAD that has input buffer (mode) enabled.
 * INPUT:
 *    pad_nb : number of the PAD (number can be found in PAD excel sheet
 *             attached to RM)
 * OUTPUT:
 *    uint8  : one bit value (0/1)
 * --------------------------------------------------------------------------*/
uint8_t uc_siul_GetGPIO(uint32_t pad_nb)
{
	return *(UC_SIUL_GPDO_BASE + pad_nb);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_swt_Disable()
 *    Disables requested Software Watchdog module.
 * INPUT:
 *    base : Requested SWT module base address in memory map
 * --------------------------------------------------------------------------*/
void uc_swt_Disable(uint32_t base)
{
	uint32_t cr_reg;

	cr_reg = *(vuint32_t*)(base + UC_SWT_CR_REG_OFFSET);
	cr_reg &= ~(uint32_t)0x01;
	*(vuint32_t*)(base + UC_SWT_SR_REG_OFFSET) = 0xC520;
	*(vuint32_t*)(base + UC_SWT_SR_REG_OFFSET) = 0xD928;
	*(vuint32_t*)(base + UC_SWT_CR_REG_OFFSET) = cr_reg;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pflash_SetPFCR1()
 *    Configuration of PLASH controller Port0.
 *    By default it enables
 *    - all masters for prefetching
 *    - data and instruction prefetch
 *    - line buffer
 * INPUT:
 *    apc    : address pipeline control
 *    ws     : read wait state control (see DS for minimum WS for system clock)
 *    pflim  : prefetch limit
 * --------------------------------------------------------------------------*/
void uc_pflash_SetPFCR1(uint32_t base, uint32_t apc, uint32_t ws, uint32_t pflim)
{
	*(vuint32_t*)(base + UC_PFLASH_PFCR1_REG_OFFSET) = UC_PFLASH_PFCR1_VAL(apc, ws, pflim);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pflash_SetPFCR2()
 *    Configuration of PLASH controller Port1.
 *    By default it enables
 *    - data and instruction prefetch
 * INPUT:
 *    pflim  : prefetch limit
 * --------------------------------------------------------------------------*/
void uc_pflash_SetPFCR2(uint32_t base, uint32_t pflim)
{
	*(vuint32_t*)(base + UC_PFLASH_PFCR2_REG_OFFSET) = UC_PFLASH_PFCR2_VAL(pflim);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pflash_SetPFCR3()
 *    Configuration of PLASH controller Port0/1
 * INPUT:
 *    p0wcfg : Port0 Way Configuration
 *    p1wcfg : Port1 Way Configuration
 *    arbm   : Arbitration Mode between ports
 * --------------------------------------------------------------------------*/
void uc_pflash_SetPFCR3(uint32_t base, uint32_t p0wcfg, uint32_t p1wcfg, uint32_t arb)
{
	*(vuint32_t*)(base + UC_PFLASH_PFCR3_REG_OFFSET) = UC_PFLASH_PFCR3_VAL(p0wcfg, p1wcfg, arb);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pram_SetPRCR1()
 *    Configuration of PRAM controller
 * INPUT:
 *    pri  : Port0/1 arbitration
 *    p0rb : Port0 read buffer enable/disable
 *    p1rb : Port1 read buffer enable/disable
 * --------------------------------------------------------------------------*/
void uc_pram_SetPRCR1(uint32_t base, uint32_t pri, uint32_t p0rb, uint32_t p1rb, uint32_t ftdis)
{
	*(vuint32_t*)(base + UC_PRAM_PRCR1_REG_OFFSET) = UC_PRAM_PRCR1_VAL(pri, p0rb, p1rb, ftdis);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_core_EnableICache()
 *    Enables Instruction cache on the current core
 * --------------------------------------------------------------------------*/
void uc_core_EnableICache(void)
{
	/* synchronize context before we access cache registers */
	__msync();
	__isync();

	/* invalidate whole cache */
    __mtspr(UC_CORE_ICACHE_SPR, (uint32_t) 0x2);
    while ((__mfspr(UC_CORE_ICACHE_SPR) & (uint32_t)0x2));

    /* enable cache */
	__mtspr(UC_CORE_ICACHE_SPR, 0x1);

	/* synchronize context after writing to cache registers */
	__isync();
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_core_EnableDCache()
 *    Enables Data cache on the current core
 * --------------------------------------------------------------------------*/
void uc_core_EnableDCache(void)
{
	/* synchronize context before we access cache registers */
	__msync();
	__isync();

	/* invalidate whole cache */
    __mtspr(UC_CORE_DCACHE_SPR, (uint32_t) 0x2);
    while ((__mfspr(UC_CORE_DCACHE_SPR) & (uint32_t)0x2));

    /* enable cache */
	__mtspr(UC_CORE_DCACHE_SPR, 0x1);

	/* synchronize context after writing to cache registers */
	__isync();
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_core_GetCurrentCore()
 *    Reads value of the current execution Core calling this function
 * OUTPUT:
 *    uint32 : Core Id
 * --------------------------------------------------------------------------*/
uint32_t uc_core_GetCurrentCore(void)
{
	uint32_t core;
	core = __mfspr(286);
	return core;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_Enable()
 *    Enables PIT module (clock).
 *    Sets Freeze in debug bit by default
 * --------------------------------------------------------------------------*/
void uc_pit_Enable(void)
{
	uint32_t mcr_reg;

	mcr_reg = UC_PIT_MCR_REG;
	mcr_reg &= ~UC_PIT_MCR_MDIS_BIT;
	mcr_reg |= UC_PIT_MCR_FRZ_BIT;
	UC_PIT_MCR_REG = mcr_reg;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_ReloadChannel()
 *    Initializes requested channel with reload value.
 * INPUT:
 *    channel      : PIT channel to initialize
 *    reload_value : period in timer ticks (PER_CLOCK)
 * --------------------------------------------------------------------------*/
void uc_pit_ReloadChannel(uint32_t channel, uint32_t reload_value)
{
	/* set reload value */
	UC_PIT_CHANNEL_REG(UC_PIT_LDVAL_REG_ADDR, channel) = reload_value;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_InitChannelIsr()
 *    Enable or disable
 * INPUT:
 *    channel      : PIT channel to initialize
 * --------------------------------------------------------------------------*/
void uc_pit_EnableChannelIsr(uint32_t channel)
{
	/* enable/disable interrupt generation in control register */
	UC_PIT_CHANNEL_REG(UC_PIT_TCTRL_REG_ADDR, channel) |= UC_PIT_TCTRL_TIE_BIT;
}

/* ----------------------------------------------------------------------------
 * * FUNCTION: uc_pit_StartCH0()
 *    Starts timer on channel 0 (the only channel supported by BSP)
 *    It programs reload value that tells a period length.
 *    User can choose if an interrupt shall be generated from the channel
 * INPUT:
 *    channel : PIT channel to initialize
 * --------------------------------------------------------------------------*/
void uc_pit_StartChannel(uint32_t channel)
{
	UC_PIT_CHANNEL_REG(UC_PIT_TCTRL_REG_ADDR, channel) |= UC_PIT_TCTRL_TEN_BIT;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_StopChannel()
 *    Stops the timer on requested PIT channel
 * INPUT:
 *    channel : PIT channel to initialize
 * --------------------------------------------------------------------------*/
void uc_pit_StopChannel(uint32_t channel)
{
	UC_PIT_CHANNEL_REG(UC_PIT_TCTRL_REG_ADDR, channel) &= ~UC_PIT_TCTRL_TEN_BIT;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_ClearChannelIsrFlag()
 *    Clear active interrupt flag on requested PIT channel
 * INPUT:
 *    channel : PIT channel to initialize
 * --------------------------------------------------------------------------*/
void uc_pit_ClearChannelIsrFlag(uint32_t channel)
{
	UC_PIT_CHANNEL_REG(UC_PIT_TFLG_REG_ADDR, channel) = UC_PIT_TFLG_TIF_BIT;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_IsChannelIsrFlag()
 *    Reads the interrupt flag
 * INPUT:
 *    channel : PIT channel to initialize
 * OUTPUT
 *    uint32 : 1 = timeout occurred
 *             0 = not yet from last clear
 * --------------------------------------------------------------------------*/
uint32_t uc_pit_IsChannelIsrFlag(uint32_t channel)
{
	return UC_PIT_CHANNEL_REG(UC_PIT_TFLG_REG_ADDR, channel);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_pit_GetChannelCurrentValue()
 *    Reads current value of the timer from requested channel
 * INPUT:
 *    channel : PIT channel to initialize
 * OUTPUT:
 *    uint32 : Timeout value
 * NOTE:
 *    timer value shall be read when timer is active, in inactive
 *    mode the value might be unreliable.
 * --------------------------------------------------------------------------*/
uint32_t uc_pit_GetChannelCurrentValue(uint32_t channel)
{
	return UC_PIT_CHANNEL_REG(UC_PIT_CVAL_REG_ADDR, channel);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_intc_SetVTBA()
 *    Register Interrupt vector table for requested Core into the INTC module.
 *    Each Core has its own vector table block in the INTC
 * INPUT:
 *    coreId : for which Core to register the vector table
 *    table  : vector table base address
 * --------------------------------------------------------------------------*/
void uc_intc_SetVTBA(uint32_t coreId, uint32_t* table)
{
	*(UC_INTC_IACKR_REG_BASE + coreId) = (uint32_t) 0xFFFFF000 & (uint32_t) table;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_intc_SetCurrentPriority()
 *    Modifies current priority setting for selected core in INTC module
 * INPUT:
 *    coreId   : for which Core to register the vector table
 *    priority : requested priority number
 * --------------------------------------------------------------------------*/
void uc_intc_SetCurrentPriority(uint32_t coreId, uint32_t priority)
{
	*(UC_INTC_CPR_REG_BASE + coreId) = (uint32_t) 0x3F & priority;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_intc_SetPSR()
 *    Assigns to selected interrupt vector priority and Core that will
 *    handle the interrupt
 * INPUT:
 *    psr_nb   : interrupt vector to configure
 *    coreId   : which core will handle the interrupt
 *    priority : requested priority level for the interrupt vector
 * --------------------------------------------------------------------------*/
void uc_intc_SetPSR(uint32_t psr_nb, uint32_t coreId, uint32_t priority)
{
	*(UC_INTC_PSR_REG_BASE + psr_nb) = UC_INTC_PSR_VAL(coreId, priority);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_intc_SetBCR()
 *    Configure Interrupt mode to use on given processor
 * INPUT:
 *    coreId   : which core will handle the interrupt
 *    mode     : 0/1 = SW/HW
 * --------------------------------------------------------------------------*/
void uc_intc_SetBCR(uint32_t coreId, uint32_t mode)
{
	if (mode != 0)
		UC_INTC_BCR_REG |= UC_INTC_BCR_VAL(coreId, mode);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: uc_intc_EndOfInterrupt()
 *    Write to INTC EOIR register to finish the External interrupt.
 *    It must be used in HW Vector mode
 * INPUT:
 *    coreId   : which core will handle the interrupt
 * --------------------------------------------------------------------------*/
void uc_intc_EndOfInterrupt(uint32_t coreId)
{
	*(UC_INTC_EOIR_REG_BASE + coreId) = 0;
}
