/******************************************************************************
 * FILE:
 *    uc_spc570s50_exceptions.S
 *
 * UC:
 *    SPC570S50
 *
 * DESCRIPTION:
 *    Core Exception table & IVOR definition and External Interrupt (IVOR4)
 *    SW handler implementation.
 *    IVOR4 is multicore capable selecting the right ITNC Core registers
 *    in the run-time.
 *
 * FILE HISTORY:
 *    V1.0    30.01.2018
 *       > Reused for spc570S50
 *    V1.1    01.03.2018     RO
 *       > Bugfix in IVOR4 handler
 *
 ******************************************************************************
 * Copyright (C) 2015-2018 HighTec EDV-Systeme GmbH.  All rights reserved.
 * This is proprietary software. Any use of the software requires a written
 * license agreement with HighTec EDV-Systeme GmbH. Please contact for
 * detailed license information: info@hightec-rt.com.
 *****************************************************************************/


/* ============================================================================
 * SYMBOLS & DEFINES & MACROS
 * ==========================================================================*/

/* Interrupt controller register addresses in memory map (uC specific)
 * IACKR0 : register to read from INTC to get an address of the handler to call
 * EOIR   : register to write 0 to signal end of Interrupt to INTC module
 * OFFSET : offset between multiple registers on multicore uC derivatives */
#define IACKR0_ADDR               0xFC040020
#define EOIR0_ADDR                0xFC040030
#define INTC_CORE_REG_OFFSET      0x04

/* Exported symbols
 * IVOR00 : base address of the Exception table */
.global IVOR00
.type IVOR00 STT_OBJECT

/* lwi: Immediate register load to avoid repetition of two instruction
 *      when 32bit value load needs to be loaded to GPR register */
lwi: .macro register, value
    e_lis  \register, \value@h
    e_or2i \register, \value@l
.endm


/* ============================================================================
 * IVOR TABLE & EXCEPTION HANDLERS
 * Low level implementation of exception handling that the user does not need
 * to care of at the beginning. Later he can implement his own exception
 * handlers that will override default handlers.
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * IVOR TABLE
 *    Table for handling the Core exceptions. It has fixed  structure
 *    and vector alignment.
 *    IVOR table is used at the very beginning of crt0 to initialize IVPR
 *    register in order to capture issues in the code soon after the reset.
 * IMPORTANT:
 *    IVOR Table needs to be align to 256B boundaries in the Linker file.
 *    That's why it has its own .section to make the placement easier.
 * --------------------------------------------------------------------------*/

	.section .ivor_table, "avx"

IVOR00: e_b   IVOR00_handler
	.align 4
IVOR01: e_b   IVOR01_handler
	.align 4
IVOR02: e_b   IVOR02_handler
	.align 4
IVOR03: e_b   IVOR03_handler
	.align 4
IVOR04: e_b   IVOR04_handler
	.align 4
IVOR05: e_b   IVOR05_handler
	.align 4
IVOR06: e_b   IVOR06_handler
	.align 4
IVOR07: e_b   IVOR07_handler
	.align 4
IVOR08: e_b   IVOR08_handler
	.align 4
IVOR09: e_b   IVOR09_handler
	.align 4
IVOR10: e_b   IVOR10_handler
	.align 4
IVOR11: e_b   IVOR11_handler
	.align 4
IVOR12: e_b   IVOR12_handler
	.align 4
IVOR13: e_b   IVOR13_handler
	.align 4
IVOR14: e_b   IVOR14_handler
	.align 4
IVOR15: e_b   IVOR15_handler

/* ----------------------------------------------------------------------------
 * Default Exception handlers
 *    Default handlers are declared as 'weak' functions to enable their
 *    overriding by the user code.
 *    They are plased in one .section to allow their easier placement
 *    in the linker file
 * --------------------------------------------------------------------------*/

	.section .ivor_handlers, "avx"
	.align    4

	.weak IVOR00_handler, IVOR01_handler, IVOR02_handler, IVOR03_handler
	.weak IVOR04_handler, IVOR05_handler, IVOR06_handler, IVOR07_handler
	.weak IVOR08_handler, IVOR08_handler, IVOR10_handler, IVOR11_handler
	.weak IVOR12_handler, IVOR13_handler, IVOR14_handler, IVOR15_handler

IVOR00_handler:	e_b       .                 /* Critical Input */
IVOR01_handler:	e_b       .                 /* Machine Check */
IVOR02_handler:	e_b       .                 /* Data Storage */
IVOR03_handler:	e_b       .                 /* Instruction Storage */
/* IVOR4: External Interrupt has a Handler below */
IVOR05_handler:	e_b       .                 /* Alignment */
IVOR06_handler:	e_b       .                 /* Program */
IVOR07_handler:	e_b       .                 /* Performance monitor */
IVOR08_handler:	e_b       .                 /* System call */
IVOR09_handler:	e_b       .                 /* Debug */
IVOR10_handler:	e_b       .                 /* EFPU Data exception */
IVOR11_handler:	e_b       .                 /* EFPU Round exception */
IVOR12_handler:	e_b       .                 /* Reserved */
IVOR13_handler:	e_b       .                 /* Reserved */
IVOR14_handler:	e_b       .                 /* Reserved */
IVOR15_handler:	e_b       .                 /* Reserved */

/* ----------------------------------------------------------------------------
 * Core(x) Shared IVOR4 External Interrupt handler
 *    Right INTC ACK and EOIR registers are selected in the handler according
 *    according to current CoreID (adds few CPU cycles to Prolog and Epilog
 *    time), but no need to have multiple copies
 *
 * STACK_FRAME (size 0x58)
 *
 * 0x58:
 * 0x54: r4 }
 * 0x50: r3 } separate copy of ISR working registers in the beg. of ISR
 * 0x4C: r12 |
 * 0x48: r11 |
 * 0x44: r10 |
 * 0x40: r9  |
 * 0x3C: r8  |
 * 0x38: r7  |
 * 0x34: r6  |
 * 0x30: r5  |
 * 0x2C: r4  |
 * 0x28: r3  |
 * 0x24: r0  | e_stmvgprw
 * 0x20: XER  ]
 * 0x1C: CTR  ]
 * 0x18: LR   ]
 * 0x14: CR   ] e_stmvsprw
 * 0x10: SRR1  >
 * 0x0C: SRR0  > e_stmvsrrw
 * 0x08:
 * 0x04:
 * 0x00: r1 (stack) for the interrupt, when stack frame is created
 * --------------------------------------------------------------------------*/

IVOR04_handler:

	/* PROLOG */
	stwu      r1, -0x58 (r1)                /* create stack frame and store back chain */
	e_stmvsrrw 0x0C(r1)                     /* save SRRs on the stack (SRR0, SRR1) */
	stw       r4, 0x54 (r1)                 /* save GPR r4 as we need a working register */
	stw       r3, 0x50 (r1)                 /* save GPR r3 as we need a working register */
	lwi       r4, IACKR0_ADDR               /* set INTC pointer to ACK registers */
	mfspr     r3, 286                       /* read CoreID SFR register */
	mulli     r3, r3, INTC_CORE_REG_OFFSET  /* get the right INTC Core ACK register (offset) */
	add       r4, r4, r3                    /* adjust the base pointer */
	lwz       r3, 0 (r4)                    /* read content of Core's INTC_IACKR register */
	wrteei    1                             /* reenable interrupts */
	e_stmvsprw 0x14 (r1)                    /* store vol. SPRs on stack (CR,LR,CTR,XER) */
	e_stmvgprw 0x24 (r1)                    /* store vol. GPRs on the stack (R0, R3-R12) */
	lwz       r3, 0x0 (r3)                  /* read vector table entry to get ISR address */
	mtlr      r3                            /* copy ISR address from r3 to lr */

	/* INTERRUPT HANDLER CALL */
	se_blrl                                 /* jump to handler */

	/* EPILOG */
	e_lmvgprw  0x24 (r1)                    /* restore vol. GPRs from stack (R0, R3-R12) */
	e_lmvsprw  0x14 (r1)                    /* restore vol. SPRs from stack (CR,LR,CTR,XER) */
	mbar      0                             /* ensure store to clear flag bit has completed */
	se_addi   r4, 0x10                      /* adjust the INTC pointer to EOIR register (0x10 offset) */
	li        r3, 0                         /* load 0 to r3 */
	wrteei    0                             /* disable interrupts */
	stw       r3, 0 (r4)                    /* write 0 to INTC_EOIR */
	lwz       r3, 0x50 (r1)                 /* restore again GPR r3 from stack */
	lwz       r4, 0x54 (r1)                 /* restore again GPR r4 from stack */
	e_lmvsrrw  0x0C (r1)                    /* restore SRRs from stack (SRR0, SRR1) */
	addi      r1, r1, 0x58                  /* delete stack frame */
	rfi                                     /* Return from Interrupt */
